[StdErr nextPutAll: '"Parsing and compiling JavaScript compiler... "]

;;; 
;;; semantics
;;; 

(define jserror (lambda (msg) (begin [StdErr nextPutAll: msg] [StdErr cr] (exit 0))))

(syntax && (lambda (node compiler) `(and ,@[node copyFrom: '1])))
(syntax || (lambda (node compiler) `(or ,@[node copyFrom: '1])))

(define LIST
  (lambda (data)
    (let ((r [WriteStream on: [Expression new: '8]]) (idx '0))
      [r nextPut: [data first]]
      (while [idx < [[data second] size]]
	(begin [r nextPut:[[[data second] at: idx] second]]
	       (set idx [idx + '1])))
      [r contents])))

(define OPTLIST (lambda (data) (if (== [[data first] size] '0) '() (LIST [data first]))))

(define PRIMEXPR
  (lambda (data)
    (let ((r [data first]) (ops [data second]) (idx '0))
      (while [idx < [ops size]]
	(let ((op [ops at: idx]))
	  (set r (cond ((== [op first] ['"[" asSymbol])	`(array-access ,r ,[op second]))
		       ((== [op first] '.)		(if (== [[op third] size] '0)
							  `(field-access ,[op second] ,r)
							  `(send ',[op second] ,r ,@[[op third] second])))
		       ((== [op first] ['"(" asSymbol])	`(,r ,@[op second]))
		       (1				(jserror '"invalid primary expression"))))
	  (set idx [idx + '1])))
      r)))

(syntax array-access		; (array-access recv idx) => [recv at: idx]
  (lambda (node compiler)
    `[,[node second] at: ,[node third]]))

(syntax array-access-set	; (array-access-set recv idx val) => [recv at: idx put: val]
  (lambda (node compiler)
    `[,[node second] at: ,[node third] put: ,[node fourth]]))
                
(syntax increment		; (increment var) => (begin (set var [var + '1]) [var - '1])
  (lambda (node compiler)
    `(begin (set ,[node second] [,[node second] + '1])
	    [,[node second] - '1])))

(syntax decrement		; (decrement var) => (begin (set var [var - '1]) [var + '1])
  (lambda (node compiler)
    `(begin (set ,[node second] [,[node second] - '1])
	    [,[node second] + '1])))

;;; 
;;; lexical analyzer
;;; 

JSSpace		::= (' ' | '\t' |  '\r' | '\n' |
		     "//" (Char - '\n')* '\n' |
		     "/*" (Char - "*/")* "*/")				=> [data first].

JSName		::= (Letter | '_') (Letter | Digit | '_' | ':')*	=> (let ((s [[[data flattened] concatenated] asSymbol]))
									     (if (or (== s 'else) (== s 'for) (== s 'function)
										     (== s 'if) (== s 'new) (== s 'return)
										     (== s 'this) (== s 'var) (== s 'while)
										     (== s 'meta))
									       `(,s)
									       `(name ,s))).

JSNumber	::= Digit+						=> `(number ,(fixNumber [[data flattened] concatenated])).

JSSpecial	::= ('(' | ')' | '{' | '}' | '[' | ']' | ',' | ';' |
		     "!=" | "==" | '=' | ">=" | '>' | "<=" | '<' |
		     "++" | '+' | "--" | '-' | '*' | '/' | '%' | "&&" |
		     "||" | '?' | '.' | '!')				=> `(,[[[data first] first] asSymbol]).

JSToken		::= JSSpace* (JSName | JSNumber | Str | JSSpecial)	=> [[data second] first].


;;; 
;;; parser
;;; 

(set Tokenizer	'JSToken)
(define Expr	0) ; forward reference
(define SrcElem	0) ; forward reference

ArgList		::= [Expr (#"," Expr)*]					=> (OPTLIST data).

Formals		::= [#name (#"," #name)*]				=> (OPTLIST data).

FuncRest	::= #"(" Formals #")" #"{" SrcElem* #"}"		=> `(lambda ,[data second] ,@[data fifth]).

PrimExprHd	::= #"(" Expr #")"					=> [data second]
		  | #name						=> [data first]
		  | #number						=> `(quote ,[data first])
		  | #string						=> `(quote ,[data first])
		  | #function FuncRest					=> [data second]
		  | #new #name #"(" ArgList #")"			=> `(send 'new ,[data second] ,@[data fourth]).

PrimExpr	::= PrimExprHd
		     (#"[" Expr #"]" |
		      #"." #name [#"(" ArgList #")"] |
		      #"(" ArgList #")")*				=> (PRIMEXPR data).

Postfix		::= PrimExpr [#"++" | #"--"]				=> (cond ((== [[data second] size] '0) [data first])
										 ((== [[data second] first] '++)
										  `(increment ,[data first]))
										 ((== [[data second] first] '--)
										  `(decrement ,[data first]))
										 (1 (jserror '"invalid postfix operator "))).

UnaryExpr	::= [#"+" | #"-" | #"!"] Postfix			=> (cond ((== [[data first] size] '0) [data second])
										 ((== [[data first] first] '+) [data second])
										 ((== [[data first] first] '-)
													`[,[data second] negated])
										 ((== [[data first] first] '!)
													`[,[data second] not])
										 (1 (jserror '"invalid unary operator"))).

(define BINEXPR
  (lambda (data wantSend)
    (let ((r [data first]) (rest [data second]) (idx '0))
      (while [idx < [rest size]]
	(begin (set r (if wantSend
			`(send ',[[rest at: idx] first] ,r ,[[rest at: idx] second])
			`(,[[rest at: idx] first] ,r ,[[rest at: idx] second])))
	       (set idx [idx + '1])))
      r)))

MultExpr	::= UnaryExpr (#"*" UnaryExpr |
			       #"/" UnaryExpr |
			       #"%" UnaryExpr)*				=> (BINEXPR data 1).

AddExpr		::= MultExpr (#"+" MultExpr |
			      #"-" MultExpr)*				=> (BINEXPR data 1).

RelExpr		::= AddExpr (#">" AddExpr |
			     #">=" AddExpr |
			     #"<" AddExpr |
			     #"<=" AddExpr)*				=> (BINEXPR data 1).

EqExpr		::= RelExpr (#"==" RelExpr |
			     #"!=" RelExpr)*				=> (BINEXPR data 0).

AndExpr		::= EqExpr (#"&&" EqExpr)*				=> (BINEXPR data 0).

OrExpr		::= AndExpr (#"||" AndExpr)*				=> (BINEXPR data 0).

Expr		::= OrExpr [#"?" Expr #":" Expr |
			    #"=" Expr]					=> (cond ((== [[data second] size] '0) [data first])
										 ((== [[data second] first] '?)
										  `(if ,[data first]
										     ,[[data second] second]
										     ,[[data second] fourth]))
										 ((== [[data second] first] '=)
										  `(set ,[data first] ,[[data second] second]))
										 (1 (jserror '"invalid top-level expression"))).

Binding		::= #name #"=" Expr					=> `(define ,[data first] ,[data third]).

Bindings	::= Binding (#"," Binding)*				=> (LIST data).

VarDecl		::= #var Bindings #";"					=> `(and ,@[data second]).

Stmt		::= #"{" Stmt* #"}"					=> `(begin ,@[data second])
		  | VarDecl						=> [data first]
		  | #if #"(" Expr #")" Stmt [#else Stmt]		=> (if (== [[data at: '5] size] '0)
									     `(if ,[data third] ,[data fifth])
									     `(if ,[data third]
									        ,[data fifth]
									        ,[[data at: '5] second]))
		  | #while #"(" Expr #")" Stmt				=> `(while ,[data third] ,[data fifth])
		  | #return [Expr] #";"					=> `(return ,@[data second])
		  | Expr #";"						=> [data first].

SrcElem		::= #function #name FuncRest				=> `(define ,[data second] ,[data third])
		  | Stmt						=> [data first].

Cmd		::= #meta Meta						=> [data second]
		  | SrcElem						=> [data first].

;;; 
;;; read-eval loop
;;; 

[StdErr nextPutAll: '"done.\n"]

;; DOM objects

(define document	(import "document"))
(define Date		(import "Date"))

[StdErr nextPutAll: '"Entering JavaScript read-eval-print loop.\n"]

(while [[OS argumentFiles] notEmpty]
  (read-eval-print Cmd [OS nextArgumentFile] 0 1 0))
